@@ -3,7 +3,7 @@
 # Accelerator for pip, the Python package manager.
 #
 # Author: Peter Odding <peter.odding@paylogic.eu>
-# Last Change: April 23, 2013
+# Last Change: April 24, 2013
 # URL: https://github.com/paylogic/pip-accel
 #
 # TODO Consider using pip's API instead of running it as a subprocess.
@@ -47,10 +47,6 @@ ENVIRONMENT = os.path.abspath(os.environ.get('VIRTUAL_ENV', sys.prefix))
 # due to connectivity issues with PyPi and/or linked distribution websites.
 MAX_RETRIES = 10
 
-# The directories under a given environment (/usr, /usr/local or a virtual
-# environment) where we can install files from binary distribution archives.
-KNOWN_ROOTS = ('bin/', 'lib/', 'man/', 'share/')
-
 # The version number of the binary distribution cache format in use. When we
 # break backwards compatibility we bump this number so that pip-accel knows it
 # should clear the cache before proceeding.
@@ -279,26 +275,26 @@ def cache_binary_distribution(input_path, output_path):
     output_bdist = tarfile.open(output_path, 'w:gz')
     for member in input_bdist.getmembers():
         # In my testing the `dumb' tar files created with the `python setup.py
-        # bdist' command contain pathnames that are relative instead of
-        # absolute, but they're relative to / which is kind of awkward: I would
-        # like to use os.path.relpath() on them but that won't give the correct
-        # result without preprocessing...
+        # bdist' command contain pathnames that are relative to `/' which is
+        # kind of awkward: I would like to use os.path.relpath() on them but
+        # that won't give the correct result without preprocessing...
         original_pathname = member.name
         absolute_pathname = re.sub(r'^\./', '/', original_pathname)
-        if member.isdir() or member.isdev():
-            debug("Ignoring directory or device file: %s\n", absolute_pathname)
-        else:
+        if member.isdev():
+            debug("Warning: Ignoring device file: %s\n", absolute_pathname)
+        elif not member.isdir():
             modified_pathname = os.path.relpath(absolute_pathname, ENVIRONMENT)
-            debug("Transformed %s -> %s.\n", original_pathname, modified_pathname)
-            if modified_pathname.startswith(KNOWN_ROOTS):
+            if os.path.isabs(modified_pathname):
+                message("Warning: Failed to transform pathname in binary distribution to relative path! (original: %r, modified: %r)\n",
+                        original_pathname, modified_pathname)
+            else:
+                debug("Transformed %r -> %r.\n", original_pathname, modified_pathname)
                 # Get the file data from the input archive.
                 file_data = input_bdist.extractfile(original_pathname)
                 # Use all metadata from the input archive but modify the filename.
                 member.name = modified_pathname
                 # Copy modified metadata + original file data to output archive.
                 output_bdist.addfile(member, file_data)
-            else:
-                message("Warning: Ignoring file in unknown location: %s\n", original_pathname)
     input_bdist.close()
     output_bdist.close()
 
@@ -332,10 +328,12 @@ def install_binary_dist(filename, install_prefix=ENVIRONMENT):
     Expects two arguments: The pathname of the tar archive and the pathname of
     the installation prefix.
     """
-    # TODO This is quite slow for modules like Django. Speed it up using links?
-    # The plan: We can maintain a "seed" environment under $PIP_ACCEL_CACHE and
-    # use symbolic and/or hard links to populate other places based on the
-    # "seed" environment.
+    # TODO This is quite slow for modules like Django. Speed it up! Two choices:
+    #  1. Run the external tar program to unpack the archive. This will
+    #     slightly complicate the fixing up of hashbangs.
+    #  2. Using links? The plan: We can maintain a "seed" environment under
+    #     $PIP_ACCEL_CACHE and use symbolic and/or hard links to populate other
+    #     places based on the "seed" environment.
     install_timer = Timer()
     python = os.path.join(install_prefix, 'bin', 'python')
     message("Installing binary distribution %s to %s ..\n", filename, install_prefix)
@@ -344,20 +342,47 @@ def install_binary_dist(filename, install_prefix=ENVIRONMENT):
         install_path = os.path.join(install_prefix, member.name)
         directory = os.path.dirname(install_path)
         if not os.path.isdir(directory):
+            debug("Creating directory: %s\n", directory)
             os.makedirs(directory)
-        debug("Writing %s\n", install_path)
+        debug("Writing file: %s\n", install_path)
         file_handle = archive.extractfile(member)
         with open(install_path, 'w') as handle:
             contents = file_handle.read()
             if contents.startswith('#!/'):
-                # Fix hashbangs.
-                # TODO Don't be so brute force about it?
-                contents = re.sub('^#![^\n]+', '#!' + python, contents)
+                contents = fix_hashbang(python, contents)
             handle.write(contents)
         os.chmod(install_path, member.mode)
     archive.close()
     message("Finished installing binary distribution in %s.\n", install_timer)
 
+def fix_hashbang(python, contents):
+    """
+    Rewrite the hashbang in an executable script so that the Python program
+    inside the virtual environment is used instead of a system wide Python.
+
+    Expects two arguments: The absolute pathname of the Python program inside
+    the virtual environment and a string with the contents of the script whose
+    hashbang should be fixed.
+
+    Returns the modified contents of the script as a string.
+    """
+    # Separate the first line in the file from the remainder of the contents
+    # while preserving the end of line sequence (CR+LF or just an LF) and
+    # without having to split all lines in the file (there's no point).
+    parts = re.split(r'(\r?\n)', contents, 1)
+    hashbang = parts[0]
+    # Get the base name of the command in the hashbang and deal with hashbangs
+    # like `#!/usr/bin/env python'.
+    modified_name = re.sub('^env ', '', os.path.basename(hashbang))
+    # Only rewrite hashbangs that actually involve Python.
+    if re.match(r'^python(\d+(\.\d+)*)?$', modified_name):
+        debug("Hashbang %r looks like a Python hashbang! We'll rewrite it!\n", hashbang)
+        parts[0] = '#!%s' % python
+        contents = ''.join(parts)
+    else:
+        message("Warning: Failed to match hashbang: %r\n", hashbang)
+    return contents
+
 def run_pip(arguments, use_remote_index):
     """
     Execute a modified `pip install` command.
